/**************************************
  Basic Disk Operating System
  CP/M v2.2 port for eZ80F92
  disk is simulated on FAT16/32 SD/MMC card
  (C) 2007 Hynek Sladky; sladky@mujmail.cz
 *************************************/

#include <stdio.h>
#include <windows.h>
#include <string.h>
#include <sys/stat.h>
#include "BIOS.h"
#include "BDOS.h"
#include "CPM_RAM.h"
#include "Common.h"

/**
ToDo:
- pokracovani v podpore dalsich FAT16/32 funkci
  o delete file
  o create file
  o write file

- podivat se na "ambiguous drive select code" v search/search next...
  asi bych to mel podporovat - coz pravdepodobne znamena hledani na vsech discich...

Poznamky k FCB:
- FCB.extent udava cislo polozky adresare (tj. polohu alokacni blok *8 nebo *16 podle poctu alokacnich bloku na disku)
- FCB.rc udava pocet sektoru v aktualnim dir_extentu;
  plny extent ma tuto hodnotu danou poctem alokacnich bloku na disku (tj. 8 nebo 16 polozek v extentu),
    a velikosti alokacniho bloku
**/

// globalni promenne
static unsigned char cbuf = 0; // char if no CTRL while checking new char
static unsigned char pout = 0; // enables output to printer
unsigned char curpos = 0; // cursor position

static unsigned short log_vec = 0x0001;
static unsigned short RO_vec = 0xFFFF; // zatim jsou vsechny disky jako R/O
static unsigned short usrDMA = 0x0080;

//static unsigned char curDiskNum = 0;
//static unsigned char curUserNum = 0;
static FCB_type *searchFCBptr;

unsigned char CheckConsole (void) {
  if (!cbuf && CONSTAT ()) {
    cbuf = CONIN ();
    if (cbuf == 0x13) {
      cbuf = CONIN ();
      if (cbuf == 0x03)
        //WBOOT (); // POZOR! Nedochazi k uvolneni zasobniku! Nutno resit jinak!!!
	longjmp (WBOOT_jmp, 0);
      cbuf = 0;
      return 0;
    }
    return 1;
  }
  return (cbuf!=0);
}

void Backspace (unsigned char c) {
  CONOUT ('\b');
  CONOUT (' ');
  BDOS (console_output, '\b');
  if (c < ' ')
    Backspace ('^');
}

void CtrlEcho (unsigned char c) {
  if (c >= ' ' || c == '\t' || c == '\r' || c == '\n' || c == '\b')
    BDOS (console_output, c);
  else {
    BDOS (console_output, '^');
    BDOS (console_output, c|0x40);
  }
}

void StartNewLine (unsigned char pos) {
  BDOS (console_output, '\r');
  BDOS (console_output, '\n');
  while (curpos < pos)
    BDOS (console_output, ' ');
}

char *makeName (char *buf, FCB_type *fptr) {
unsigned char i;
char *s;
  s=buf;
  for (i=0; i<11; i++) {
    if (fptr->name[i]==' ') {
      i=(i<8)?7:10;
      continue;
    }
    if (i==8)
      *buf++='.';
    *buf++=fptr->name[i];
  }
  *buf=0;
  return s;
}

int fillFCB (FCB_type *fptr, char *buf);

unsigned long makefilepos (unsigned char modnum, unsigned char extent, unsigned char currec) {
unsigned long pos;
  pos = modnum & 0x0F;
  pos <<= 5;
  pos |= extent & 0x1F;
  pos <<= 7;
  pos |= currec & 0x7F;
  pos <<= 7;
  return pos;
}

extern char BDOS_verbose;

unsigned short BDOS (unsigned char func, unsigned short par) {
unsigned char tmpc;
unsigned char cntb;
unsigned char *ptr;
unsigned char startpos;
static WIN32_FIND_DATA dir; // static, abych ji mohl sdilet mezi vice volanima...
static char name[16];
int i;
unsigned long tmpl;
FCB_type *fptr;
struct stat stat;

  switch (func) {
    // --------------------------------
    // BOOT
    // --------------------------------
    case system_reset:
      longjmp (BOOT_jmp, 0);
    // --------------------------------
    // cteni znaku s echem
    // --------------------------------
    case console_input:
      tmpc = (cbuf)?cbuf:CONIN ();
      cbuf = 0;
      // kontrola na CR, LF, TAB, BS
      if (tmpc >= ' ' || tmpc == '\t' || tmpc == '\r' || tmpc == '\n' || tmpc == '\b') 
        BDOS (console_output, tmpc);
      return tmpc;
    // --------------------------------
    // vystup znaku s mirnou podporou pro funkci read_console_buffer
    // --------------------------------
    case console_output:
      // tabulator
      if (((unsigned char)par) == '\t') {
        do {
	  BDOS (console_output, ' ');
	} while (curpos & 0x07);
	break;
      }
      // pokud se stiskne ^S a pak ^C, mel by se system restartovat: WBOOT
      //if (!cout) {
        CheckConsole ();
	CONOUT (par);
        // pokud je povoleny soucasny tisk (povoluje se ve funkci 10), tiskne se kazdy znak
	if (pout) LIST (par);
      //}
      // pozice kurzoru
      if (((unsigned char)par) >= ' ' && ((unsigned char)par) != 0x7F) curpos++;
      if (curpos && ((unsigned char)par) == '\b') curpos--;
      if (((unsigned char)par) == '\n') curpos = 0;
      break;
    // --------------------------------
    // tape reader input; used for serial input
    // --------------------------------
    case reader_input:
      return READER ();
    // --------------------------------
    // tape punch output; used for serial output
    // --------------------------------
    case punch_output:
      PUNCH (par);
      break;
    // --------------------------------
    // printer output
    // --------------------------------
    case list_output:
      LIST (par);
      break;
    // --------------------------------
    // direct console I/O through BDOS
    // --------------------------------
    case console_direct:
      if (((unsigned char)par) == 0xFF) {
        if (CONSTAT ())
	  return CONIN ();
	break;
      }
      if (((unsigned char)par) == 0xFE)
        return CONSTAT ();
      CONOUT (par);
      break;
    // --------------------------------
    // ioByte get/set
    // --------------------------------
    case getIObyte:
      return ioByte;
    case setIObyte:
      ioByte = par;
      break;
    // --------------------------------
    // print string ended with '$'
    // --------------------------------
    case print_string:
      while (CPM_RAM[par]!='$')
        CONOUT (CPM_RAM[par++]);
      break;
    // --------------------------------
    // read console and return whole line in buffer
    // --------------------------------
    case read_console_buffer:
      // CPM_RAM[par] = delka bufferu
      // CPM_RAM[par+1] = aktualni delka dat
      // CPM_RAM[par+2] = prvni znak...
      ptr = &CPM_RAM[par+2];
      cntb = 0;
      startpos = curpos;
      do {
        // precteni znaku
        tmpc = (cbuf)?cbuf:CONIN ();
	cbuf = 0;
	//tmpc &= 0x7F;
	// LF by se na zacatku radku melo potlacit
	if (tmpc=='\n' && !cntb) continue;
	// CR/LF konci zadavani
	if (tmpc == '\r' || tmpc == '\n')
	  break;
	// BS maze znak, na zacatku radky ignorovan
        if (tmpc == '\b') {
          if (cntb) {
	    cntb--;
	    ptr--;
	    Backspace (*ptr);
	  }
	  continue;
	}
	// DEL smaze znak a vytiskne posledni zadany
	if (tmpc == 0x7F) {
          if (cntb) {
	    cntb--;
	    ptr--;
	    // echo predchoziho znaku
            CtrlEcho (*ptr);
	  }
	  continue;
	}
	// ^E posune tisk znovu od zacatku
	if (tmpc == 0x05) { // ^E
	  StartNewLine (startpos);
	  continue;
	}
	// ^P zapne/vypne tisk na tiskarnu
	if (tmpc == 0x10) { // ^P
	  pout = !pout; //??
	  continue;
	}
	// ^X smaze cely radek
        if (tmpc == 0x18) {
	  while (cntb) {
	    cntb--;
	    ptr--;
	    Backspace (*ptr);
	  }
	  continue;
	}
	// ^U prejde na novy radek a smaze buffer
	if (tmpc == 0x15) {
          BDOS (console_output, '#');
	  StartNewLine (startpos);
	  cntb = 0;
	  continue;
	}
        // ^R znovu vypise zadanou radku
	if (tmpc == 0x12) {
          BDOS (console_output, '#');
	  StartNewLine (startpos);
          ptr = &CPM_RAM[par+2];
	  for (tmpc = cntb; tmpc; tmpc--) {
	    CtrlEcho (*ptr);
	    ptr++;
	  }
	  continue;
	}
        // ^C skoci na WBOOT, pokud je zacatek radku
	if (tmpc == 0x03 && !cntb)
	  longjmp (WBOOT_jmp, 0);
	// add char to buffer
	*ptr = tmpc;
	cntb++;
	ptr++;
	// echo
	CtrlEcho (tmpc);
      } while (cntb < CPM_RAM[par]);
      CPM_RAM[par+1]=cntb;
      BDOS (console_output, '\r');
      break;
    // --------------------------------
    // get console status
    // --------------------------------
    case get_console_status:
      return CheckConsole ();
    // --------------------------------
    // get CP/M version number
    // --------------------------------
    case version_number:
      return 0x0022;
    // --------------------------------
    // reset all disks
    // --------------------------------
    case reset_disk_system:
      if (BDOS_verbose) dprintf ("BDOS: reset disks\n");
      RO_vec = 0;
      log_vec = 1;
      SELDSK (0);
      break;
    // --------------------------------
    // select disk
    // --------------------------------
    case select_disk:
      if (BDOS_verbose) dprintf ("BDOS: select disk %02X\n", par);
      SELDSK (par);
      break;
    // --------------------------------
    // open file
    // --------------------------------
    case open_file:
      // teoreticky bych nemel povolit nejednoznacne jmeno... ale original to taky nekontroluje...
      SETDMA (dirbuf_offs);
      tmpc = BDOS (search_first, par);
      SETDMA (usrDMA);
      if (tmpc) return 0xFF;
      // close directory search file
      fptr = (FCB_type*)(CPM_RAM+par);
      FindClose (fptr->f);
      // save current disk
      tmpc = curDisk&0x0F;
      if (fptr->drive && (fptr->drive-1)!=(curDisk&0x0F))
        SELDSK (fptr->drive-1);
      fptr->f = fopen (makeName (name, fptr), "rb+");
      // restore disk
      if (tmpc != (curDisk & 0x0F))
        SELDSK (tmpc);
      if (!fptr->f)
	return 0xFF;
      if (BDOS_verbose) dprintf ("BDOS: open file %04X: %s [%02X %02X %02X]\n", par, name, fptr->modnum, fptr->extent, fptr->currec);
      // initialize variables
      // modnum:extent:currec udavaji pozadovanou pozici pro otevreni
      // podle bdos.asm se zachova extent, modnum je pravdepodobne 0, currec se necha nastaveny
      fptr->modnum = 0x80; // position (hi) <0..15>
//      fptr->extent = 0; // position (med) <0..31>
      fptr->currec = 0; // position (lo) <0..127>
      // podle pozadovane polohy nastavim pocet sektoru daneho extentu
      fstat (fileno (fptr->f), &stat);
      i = (stat.st_size - makefilepos (fptr->modnum, fptr->extent, 0)) >> 7;
      fptr->reccount = (i>128) ? 128 : i;
      break;
    // --------------------------------
    // close file
    // I am afraid of non closing files for reading...
    // --------------------------------
    case close_file:
      if (BDOS_verbose) dprintf ("BDOS: close file %04X\n", par);
      fptr = (FCB_type*)(CPM_RAM+par);
      if (!(fptr->modnum&0x80)) dprintf ("BDOS: Possible read-write request!\n");
      fclose (fptr->f);
      fptr->f = 0;
      break;
    // --------------------------------
    // search file on disk
    // --------------------------------
    case search_first:
      searchFCBptr = (FCB_type*)&CPM_RAM[par];
      if (BDOS_verbose) dprintf ("BDOS: search first %04X %s\n", par, makeName (name, searchFCBptr));
      searchFCBptr->f=0;
    // --------------------------------
    // search necht file on disk
    // --------------------------------
    case search_next:
      if (BDOS_verbose) dprintf ("BDOS: search next %04X -> %04X\n", par, DMA);
      do {
	if (!searchFCBptr->f) {
	  // save current disk
	  tmpc = curDisk&0x0F;
	  if (searchFCBptr->drive!='?' && searchFCBptr->drive && (searchFCBptr->drive-1)!=(curDisk&0x0F))
	    SELDSK (searchFCBptr->drive-1);
	  searchFCBptr->f = FindFirstFile (makeName (name, searchFCBptr), &dir);
	  // restore disk
	  if (tmpc != (curDisk & 0x0F))
	    SELDSK (tmpc);
	} else {
	  if (!FindNextFile (searchFCBptr->f, &dir))
	    break;
	}
	if (searchFCBptr->f==INVALID_HANDLE_VALUE) return 0xFF;
        if (dir.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) continue;
	ptr = *dir.cAlternateFileName?dir.cAlternateFileName:dir.cFileName;
	str_up (ptr);
	fillFCB ((FCB_type*)name, ptr);
        // porovnani se zadanou maskou
	for (cntb=0; cntb<11; cntb++) {
	  tmpc = searchFCBptr->name[cntb];
	  if (tmpc>='a' && tmpc<='z') tmpc &= ~0x20;
	  if (tmpc!='?' && tmpc!=name[1+cntb])
	    break;
	}
	if (cntb>=11) {
	  // nakopirovani polozky adresare na adresu DMA (32B)
	  memset (&CPM_RAM[DMA], 0, 16);
	  // CatChum.com needs (checks for) cluster numbers in directory entry!
	  memset (&CPM_RAM[DMA+16], 0x5A, 16); // fake cluster numbers
	  memcpy (&CPM_RAM[DMA+1], name+1, 11);
	  if (dir.dwFileAttributes & FILE_ATTRIBUTE_READONLY) CPM_RAM[DMA+9] |= 0x80;
	  if (dir.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) CPM_RAM[DMA+10] |= 0x80;
	  if (dir.dwFileAttributes & FILE_ATTRIBUTE_ARCHIVE) CPM_RAM[DMA+11] |= 0x80;
	  if (BDOS_verbose) dprintf ("BDOS: search OK\n");
	  return 0;
	}
      } while (1);
      FindClose (searchFCBptr->f);
      searchFCBptr->f=0;
      return 0xFF;
    // --------------------------------
    // delete file
    // --------------------------------
    case delete_file:
      // maze vsechny soubory odpovidajici danemu FCB
      fptr = (FCB_type*)(CPM_RAM+par);
      if (BDOS_verbose) dprintf ("BDOS: delete file %04X: %s\n", par, makeName (name, fptr));
      // save current DMA buffer and disk
      SETDMA (dirbuf_offs);
      tmpc = curDisk&0x0F;
      if (fptr->drive && (fptr->drive-1)!=tmpc)
        SELDSK (fptr->drive-1);
      cntb = BDOS (search_first, par);
      i=-1;
      while (!cntb) {
	i = remove (makeName (name, fptr));
	if (i) break;
	cntb = BDOS (search_next, par);
      }
      // restore DMA buffer and disk
      SETDMA (usrDMA);
      if (tmpc != (curDisk & 0x0F))
        SELDSK (tmpc);
      if (!cntb) FindClose (fptr->f);
      return (i)?0xFF:0;
    // --------------------------------
    // random read file
    // --------------------------------
    case file_read_random:
      fptr = (FCB_type*)(CPM_RAM+par);
      if (BDOS_verbose) dprintf ("BDOS: seek file %04X (%04X)\n", par, fptr->recreq);
      //FileSeek (fptr->f, fptr->recreq << 7);
      fptr->currec = fptr->recreq & 0x7F;
      fptr->extent = (fptr->recreq >> 7) & 0x1F;
      fptr->modnum &= 0xE0;
      fptr->modnum |= (fptr->recreq >>12) & 0x0F;
      tmpl = fptr->recreq << 7;
      goto Seek4Read;
    // --------------------------------
    // sequential file read
    // --------------------------------
    case read_file:
      fptr = (FCB_type*)(CPM_RAM+par);
      tmpl = makefilepos (fptr->modnum, fptr->extent, fptr->currec);
    Seek4Read:
      if (tmpl != ftell (fptr->f))
        fseek (fptr->f, tmpl, SEEK_SET);
      if (BDOS_verbose) dprintf ("BDOS: read file %04X (%lX) -> %04X [%02X %02X %02X]:%lX\n", par, ftell (fptr->f), DMA, fptr->modnum, fptr->extent, fptr->currec, tmpl);
      i = fread (CPM_RAM+DMA, 1, 128, fptr->f);
      if (i<0) return 6;
      if (!i) return 1;
      while (i<128) {
        CPM_RAM[DMA+i]='\x1A';
	i++;
      }
      fptr->currec++;
      if (fptr->currec<128) break;
      fptr->currec=0;
      fptr->extent++;
      if (fptr->extent>=32) {
        fptr->extent=0;
        fptr->modnum++;
      }
      // modnum obsahuje stavove bity 7 a 6... ???
      if ((fptr->modnum&0x3F)>15) return 6;
      // recalc extent size - record count
      fstat (fileno (fptr->f), &stat);
      i = (stat.st_size - makefilepos (fptr->modnum, fptr->extent, 0)) >> 7;
      fptr->reccount = (i>128) ? 128 : i;
/*      if (BDOS_verbose) {
        for (cntb=0; cntb<128; cntb++)
	  //dprintf ("%02X ", CPM_RAM[DMA+cntb]);
	  dputc (CPM_RAM[DMA+cntb]);
        dputc ('\n');
      }*/
      break;
    // --------------------------------
    // random write file
    // --------------------------------
    case file_write_random:
      fptr = (FCB_type*)(CPM_RAM+par);
      if (BDOS_verbose) dprintf ("BDOS: seek file %04X (%04X)\n", par, fptr->recreq);
      //FileSeek (fptr->f, fptr->recreq << 7);
      fptr->currec = fptr->recreq & 0x7F;
      fptr->extent = (fptr->recreq >> 7) & 0x1F;
      fptr->modnum &= 0xE0;
      fptr->modnum |= (fptr->recreq >>12) & 0x0F;
      tmpl = fptr->recreq << 7;
      goto Seek4Write;
    // --------------------------------
    // sequential file write
    // --------------------------------
    case write_file:
      fptr = (FCB_type*)(CPM_RAM+par);
      tmpl = makefilepos (fptr->modnum, fptr->extent, fptr->currec);
    Seek4Write:
      if (tmpl != ftell (fptr->f))
        fseek (fptr->f, tmpl, SEEK_SET);
      if (BDOS_verbose) dprintf ("BDOS: write file %04X (%lX) -> %04X [%02X %02X %02X]:%lX\n", par, ftell (fptr->f), DMA, fptr->modnum, fptr->extent, fptr->currec, tmpl);
      if (BDOS_verbose) {
        for (cntb=0; cntb<128; cntb++)
	  //dprintf ("%02X ", CPM_RAM[DMA+cntb]);
	  dputc (CPM_RAM[DMA+cntb]);
        dputc ('\n');
      }
      i = fwrite (CPM_RAM+DMA, 1, 128, fptr->f);
      //if (i<0) return 6;
      if (!i) return 5;
      fptr->currec++;
      if (fptr->currec<128) break;
      fptr->currec=0;
      fptr->extent++;
      if (fptr->extent>=32) {
        fptr->extent=0;
        fptr->modnum++;
      }
      // modnum obsahuje stavove bity 7 a 6... ???
      if ((fptr->modnum&0x3F)>15) return 6;
      // recalc extent size - record count
      fstat (fileno (fptr->f), &stat);
      i = (stat.st_size - makefilepos (fptr->modnum, fptr->extent, 0)) >> 7;
      fptr->reccount = (i>128) ? 128 : i;
      break;
    // --------------------------------
    // create new file
    // --------------------------------
    case make_file:
      fptr = (FCB_type*)(CPM_RAM+par);
      if (BDOS_verbose) dprintf ("BDOS: make file %04X: %s\n", par, makeName (name, fptr));
      // save current disk
      tmpc = curDisk&0x0F;
      if (fptr->drive && (fptr->drive-1)!=tmpc)
        SELDSK (fptr->drive-1);
      fptr->f = fopen (name, "wb+");
      // restore disk
      SETDMA (usrDMA);
      if (tmpc != (curDisk & 0x0F))
        SELDSK (tmpc);
      return (fptr->f)?0:0xFF;
    // --------------------------------
    // rename file
    // --------------------------------
    case rename_file:
      illegal ("rename file");
      break;
    // --------------------------------
    // return disk login vector
    // --------------------------------
    case login_vector:
      return log_vec;
    // --------------------------------
    // get current disk
    // --------------------------------
    case current_disk:
      return curDisk & 0x0F;
    // --------------------------------
    // set DMA address
    // --------------------------------
    case setDMA:
      usrDMA = par;
      SETDMA (par);
      break;
    // --------------------------------
    // get disk allocation vector
    // --------------------------------
    case get_alloc_vector:
      //illegal ("get ALV");
      return ALV_offs;
    // --------------------------------
    // write protect disk
    // --------------------------------
    case write_protect_disk:
      RO_vec |= (1 << (curDisk & 0x0F));
      break;
    // --------------------------------
    // get read-only disk vector
    // --------------------------------
    case getROvect:
      return RO_vec;
    // --------------------------------
    // set file attributes
    // --------------------------------
    case file_set_attr:
      illegal ("file attr");
      break;
    // --------------------------------
    // get disk parameters block
    // --------------------------------
    case get_diskpar:
      //illegal ("get disk param");
      return DPB_offs;
    // --------------------------------
    // get/set user code
    // --------------------------------
    case get_set_user:
      if ((par&0xFF)==0xFF)
        return curDisk>>4;
      curDisk = (curDisk&0x0F) | (unsigned char)(par<<4);
      break;
    // --------------------------------
    // compute file size
    // note: file is closed!
    // second note: CP/M allows partial block allocation while random written; only block really written are allocated for the file!
    // --------------------------------
    case file_size:
      if (BDOS_verbose) dprintf ("BDOS: get file size %04X\n", par);
      SETDMA (dirbuf_offs);
      tmpc = BDOS (search_first, par);
      SETDMA (usrDMA);
      if (tmpc) return 0xFF;
      fptr = (FCB_type*)(CPM_RAM+par);
      // close directory search file
      FindClose (fptr->f);
      fptr->recreq = dir.nFileSizeLow >> 7;
      break;
    // --------------------------------
    // compute random position from sequential position
    // --------------------------------
    case set_random_record:
      if (BDOS_verbose) dprintf ("BDOS: get random pos %04X\n", par);
      // spravne bych to mel pocitat z modnum:extent:currec...
      ((FCB_type*)(CPM_RAM+par))->recreq = ftell (((FCB_type*)(CPM_RAM+par))->f) >> 7;
      break;
    // --------------------------------
    // reset selected drive
    // --------------------------------
    case reset_drive:
      if (BDOS_verbose) dprintf ("BDOS: reset drive %04X\n", par);
      RO_vec &= ~par;
      log_vec &= ~par;
      break;
    // --------------------------------
    // random write with filling all unallocated block with zeros
    // --------------------------------
    case file_write_random_zero:
      illegal ("fill/random write");
      break;

//    default:
//      return 0;
  }
  return 0;
}

void illegal (char *info) {
  dprintf ("\nIllegal call: %s! System halted.", info);
  while (CONIN() != 0x03);
  longjmp (BOOT_jmp, 0);
}

