/**************************************
  Basic Input Output System
  CP/M v2.2 port for eZ80F92
  disk is simulated on FAT16/32 SD/MMC card
  (C) 2007 Hynek Sladky; sladky@mujmail.cz
 *************************************/

#include <stdio.h>
#include <windows.h>

#include "CPM_RAM.h"
#include "z80\z80.h"

#include "BIOS.h"
#include "BDOS.h"
#include "CCP.h"

/**
ToDo:
o jednoduche funkce
  SETDMA
x logicke diskove funkce
  SELDSK
x fyzicke diskove funkce? Asi nebudou implementovany...
  HOME
  SETTRK
  SETSEC
  READ
  WRITE
  SECTRN

- pro ladeni asi bude stejne nejlepsi dat CPM_RAM jako ROM, protoze se naprogramuje i kod do TPA...
  ted mam zazalohovanou verzi pro RAM a copy to ROM, tak zkusim prenastavit na ROM only...

- stale mi neni jasne, co to je submit file (autoexec) a jak se to spousti...

- inicializace CPM_RAM kodem z ROM
- inicializace jednotlivych skoku (BIOS, BDOS, BDOS, ...)
  mozna bude potrebovat uvest XDEF pro jednotlive adresy

o rozdil mezi BOOT a WBOOT?
o kontrolovat stack pri navratu z Z80 rezimu?
  resp. zajistit, aby stack nerostl nade vsechny meze...

- optimalizovat cely system:
  - prevest funkce UART1 primo do odpovidajicich funkci BIOS
    tj. uplne zrusit UART1, pro FAT a SPI vytvorit nahradni UART1.h, 
      ktery nacte BIOS.h a premapuje dputc, dputs a dprintf na BIOS funkce
      tj. BIOS musi obsahovat ekvivalent dprintf
    konec koncu muzu do BIOS.h napsat define pro propojeni funkci UART1.h a BIOS.h???
    #define CONOUT dputc
    asi bude nutne upravit parametry, respektive jejich rozmer, a navratovou hodnotu dkbhit... (0xFF)
- funkce pro primy pristup na disk by asi mely zatim psat nejakou hlasku na terminal
  a pak se bud vratit nebo zustat v nekonecne smycce...
o priprava pameti
  - namapovani 64KB RAM na konec volne RAM
  - priprava funkci pro:
    - spusteni Z80 kodu (spusteni nebo navrat)
    - skok do BIOS/BDOS a priprava parametru na zasobnik
    - vymyslet, jak realizovat tabulku skoku na funkce BIOSu
      asi by bylo rozumne, abych k tomu napsal asm kod, ktery se na uvedene adresy RAM vlozi pri BOOT
    - testovaci program v CP/M - psany v asm a zkopirovany na adresu 0x0100 pameti
  -> k tomu si musim prinest eZ80 CPU manual!!!
**/
// ====================================
// variables
// ====================================
unsigned short DMA;
jmp_buf WBOOT_jmp;
static char dirs[DISK_CNT][1024];

// ====================================
// general functions
// ====================================
// cold boot
// ------------------------------------
// inicializace uplne vseho HW i SW
// pri spusteni CCP obsahuje C registr 0x00 - tj. disk A
// WBOOT necha mozna vybraneho uzivatele a disk?
// hlavne WBOOT pocita s tim, ze uzivatelsky program mohl prepsat CCP a BDOS, tak je nahraje a zinicializuje
// asi to udelam tak, ze main=BOOT provede inicializaci HW, pak se zavola WBOOT, ktery provede reinicializaci SW a spusti CCP
//void BOOT (void) {
extern char sys_path[];
void BOOT (void) {
int i;
  // tisk uvodniho hlaseni
  dputs ("\n\f\n64k CP/M vers 2.2");
  dputs ("COPYRIGHT (C) 1979, DIGITAL RESEARCH");
  // eZ80 CP/M: copyright portovaneho systemu
  dputs ("port for x86 and FAT16/32 (C) 2007 Hynek Sladky");
  // eZ80 CP/M: inicializace SD/MMC karty

  // inicializace RAM
  for (i=0; i<0x10000; i++)
    CPM_RAM[i]=0;
  
  for (i=0; i<DISK_CNT; i++) 
    *dirs[i]=0;
  // nastaveni aktualnich adresaru
  for (i=0; i<DISK_CNT; i++)
    strcpy (dirs[i], sys_path);
  // nastaveni disku A
  curDisk = 0;
  SetCurrentDirectory (dirs[curDisk]);
  // nastaveni ioByte, pokud je implementovano
  ioByte = 0;
  // -> takze zatim neni a ani neprodpokladam, ze bude
  // spusteni CP/M {gocpm}
  WBOOT ();
}
// ====================================
// warm boot
// ------------------------------------
// CP/M se nahraje z disku A do RAM
// nastaveni RAM: skoky na BIOS a BDOS, IOBYTE, USER/DISK
// zaroven ma cist USER/DISK a nastavit posledni vybrany disk do C pro CCP
void WBOOT (void) {
const unsigned short dph[8]={0,0,0,0,dirbuf_offs,DPB_offs,CSV_offs,ALV_offs};
const DPB_type dpb={65535,7,127,7,65535,512,0x80,0x00,0,0};
int i;

  // pro nestandardni pripady lze skakat sem:
  setjmp (WBOOT_jmp);
  // inicializace FAT
  // uvolneni vsech otevrenych souboru
  // krome SUMBIT, pokud ho budu podporovat...
//  FAT_close_all ();
  // inicializace HW
  // defaultni buffer pro diskove operace je CPM_RAM[0x0080]
  DMA = defaultBUF - CPM_RAM;
  // priprava CP/M RAM
  Z80state.ram = (unsigned long) CPM_RAM;
  // reset
  Z80state.iff1=0;
  Z80state.iff2=0;
  Z80state.im=1;
  Z80state.pc=0;
  // nastaveni obsahu RAM:
  // 0x0000 = skok na wboote
  CPM_RAM[0] = 0xC3;
  *((unsigned short*)(&CPM_RAM[1])) = (unsigned short)wboot_vec;
  // 0x0005 = skok na bdos
  CPM_RAM[5] = 0xC3;
  *((unsigned short*)(&CPM_RAM[6])) = (unsigned short)bdos_vec;
  // skok na BDOS trap
  CPM_RAM[bdos_vec] = 0xC3;
  *((unsigned short*)(&CPM_RAM[bdos_vec+1])) = (unsigned short)bdos_trap;
  // BDOS trap
  CPM_RAM[bdos_trap] = 0x76; // HALT for BDOS func
  CPM_RAM[bdos_trap+1] = 0xC9; // RET from BDOS func
  // tabulka skoku na BIOS traps a BIOS traps
  for (i=0; i<17; i++) {
    CPM_RAM[wboot_vec-3+i*3] = 0xC3;
    *((unsigned short*)(&CPM_RAM[wboot_vec-2+i*3])) = boot_trap-2*i;
    CPM_RAM[boot_trap-2*i] = 0x76; // HALT for BIOS func
    CPM_RAM[boot_trap-2*i+1] = 0xC9; // RET from BIOS func
  }
  // nastaveni skoku v BIOS oblasti
  // 0x0038 = skok na mon80 ???
  CPM_RAM[0x0038] = 0xC9;
  // nastaveni datovych struktur disku
  memcpy (&CPM_RAM[diskParHeader_offs], dph, sizeof (dph));
  memcpy (&CPM_RAM[DPB_offs], &dpb, sizeof (dpb));

  // C registr <= posledni nalogovany disk (cdisk)
  // spusteni CCP {cpmb}
  CCP ();
}
// ====================================
// console functions
// ====================================
// console state
// ------------------------------------
/**
unsigned char CONST (void) {
  return dkbhit ()?0xFF:0x00;
}
**/
// ====================================
// console input
// ------------------------------------
/**
unsigned char CONIN (void) {
  return dgetc ();
}
**/
// ====================================
// console output
// ------------------------------------
/**
void CONOUT (unsigned char c) {
  dputc (c);
}
**/
// ====================================
// additional devices' functions
// ====================================
// printer state
// ------------------------------------
unsigned char LISTST (void) {
  return 0;
}
// ====================================
// printer output
// ------------------------------------
void LIST (unsigned char c) {
}
// ====================================
// punch one byte
// ------------------------------------
void PUNCH (unsigned char c) {
}
// ====================================
// read one punched byte
// ------------------------------------
unsigned char READER (void) {
  return 0;
}
// ====================================
// disk functions
// ====================================
// set home position (track 0) on active disk
// ------------------------------------
void HOME (void) {
  illegal ("HOME");
}
// ====================================
// set requested disk
// return disk parameter block header
// ------------------------------------
unsigned short SELDSK (unsigned char disk) {
  if (disk>=DISK_CNT) return 0;
  if ((curDisk & 0x0F) != disk) {
    GetCurrentDirectory (1024, dirs[curDisk & 0x0F]);
    curDisk = (curDisk & 0xF0) | disk;
    SetCurrentDirectory (dirs[disk]);
  }
  return diskParHeader_offs;
}
// ====================================
// set track
// ------------------------------------
void SETTRK (unsigned short track) {
  illegal ("SET TRK");
}
// ====================================
// set sector
// ------------------------------------
void SETSEC (unsigned short sector) {
  illegal ("SET SEC");
}
// ====================================
// set DMA address
// ------------------------------------
void SETDMA (unsigned short addr) {
  DMA = addr;
}
/**************************************
BIOS disk limits:
- original BIOS saves track and sector as 8bit value
  256 * 256 * 128 = 8MB disk size
- parameters allow to enter 16bit values
  theoretical maximum disk size is
  65536 * 65536 * 128 = 512GB disk size!!!
BDOS has another limitations:
- maximum block size is 128*128 = 16KB
- for 8bit block number, maximum accessible disk size is
  256 * 16KB = 4MB
- for 16bit block number
  65536 * 16K = 1GB
BDOS limitation for file size:
- record number is 16bit
  65536 * 128 = 8MB file size
- allocation block limit (16bit block number)
  8 allocation block numbers in one directory entry
  256 directory entry extensions
  -> totally 2048 allocation blocks for one file
  => for 16KB allocation block: 32MB file size
**************************************/
// ====================================
// read requested sector
// ------------------------------------
unsigned char READ (void) {
  illegal ("SEC READ");
  return 1;
}
// ====================================
// write requested sector
// ------------------------------------
unsigned char WRITE (void) {
  illegal ("SEC WRITE");
  return 1;
}
// ====================================
// convert logic sector number to physical sector number
// ------------------------------------
unsigned short SECTRN (unsigned short sector, unsigned short table) {
  return sector;
}

