/**************************************
  Console Command Processor
  CP/M v2.2 port for eZ80F92
  disk is simulated on FAT16/32 SD/MMC card
  (C) 2007 Hynek Sladky; sladky@mujmail.cz
 *************************************/

#include <stdio.h>
#include <windows.h>
#include <string.h>
#include "z80\z80.h"
//#include "BIOS.h"
#include "BDOS.h"
#include "CCP.h"
#include "CPM_RAM.h"
// ciste teoreticky bych mel pouzivat jen BDOS funkce, ale protoze chci podporovat vic moznosti FAT16/32, pouzivam take FAT knihovnu
//#include "FAT.h"
//#include "SPI.h"
#include "Common.h"

/**
ToDo:
- rozhodnout se, zda ma byt CCP postaven na prenosnoti smerem k nizsim nebo vyssim systemum
  - bud bude CCP pouzivat jen funkce BDOS/BIOS -> prenositelnost k nizsim systemum
  - nebo bude CCP pouzivat jen funkce vyssich systemu (stdio/FAT apod)
- naprogramovat CCP vestavene funkce kompatibilni s CP/M 2.2
  zakladni funkce museji byt plne kompatibilni s CP/M 2.2!
  - rozsirene funkce se museji spoustet bud s nejakym parametrem, nebo museji mit jiny nazev...
  -> nastudovat CCP funkce a jejich parametry
o volani uzivatelskeho programu je provedeno pouze "call 0x0100", pote je nastaveny sp na hodnotu vhodnou pro CCP
  takze asi bude dobre nastavit SP na nejakou zajimavou adresu, treba pred BDOS misto CCP...

**/

// ====================================
// CCP global variables
// ------------------------------------
unsigned int freq = 10000;
extern int BREAK_request;
// ====================================
// command structures and functions
// ------------------------------------
enum {
  cmd_help, 
  cpm_dir, cpm_era, cpm_ren, cpm_type, cpm_save, cpm_user,
  cmd_reset, cmd_cd,
  cmd_dir, cmd_type,
  cmd_tpa, cmd_bflags, cmd_exit, cmd_freq, cmd_num};
static const struct {
  const char *cmd;
  int code;
} commands[cmd_num] = {
  {"dir", cpm_dir},
  {"era", cpm_era},
  {"ren", cpm_ren},
  {"type", cpm_type},
  {"save", cpm_save},
  {"user", cpm_user},
  {"help", cmd_help},
  {"reset", cmd_reset},
  {"xdir", cmd_dir},
  {"xtype", cmd_type},
  {"cd", cmd_cd},
  {"tpa", cmd_tpa},
  {"bflg", cmd_bflags},
  {"exit", cmd_exit},
  {"freq", cmd_freq}
};
static int FindCommand (char *cmd) {
int i, j;
  for (i=0; i<cmd_num; i++) {
    if (!strnicmp (cmd, commands[i].cmd, j=strlen (commands[i].cmd)))
      //if (cmd[j]>='\0' && cmd[j]<=' ')
      if (cmd[j]<=' ')
        //return commands[i].code;
	return i;
  }
  return -1;
}
// ====================================
// directory listing functions
// ------------------------------------
// mozne formaty vystupu:
// - datum cas atributy LFN
// - [short] [datum cas] [atributy] [velikost] [LFN]
#define DIR_show_short      0x0001
#define DIR_show_time       0x0002
#define DIR_show_attr       0x0004
#define DIR_show_size       0x0008
#define DIR_show_LFN        0x0010

#define DIR_show_internals  0x0100
#define DIR_show_dot        0x0200
#define DIR_force_short     0x0400
#define DIR_column          0x0800
// ====================================
// print filename based on request flags and current directory entry
// ------------------------------------
static int PrintFileName (WIN32_FIND_DATA *dir, unsigned int flags) {
int i, j;
char *s;
  if (*dir->cAlternateFileName && !(flags & DIR_force_short)) return dprintf ("%s", dir->cFileName);
  i=0;
  s=(*dir->cAlternateFileName)?dir->cAlternateFileName:dir->cFileName;
  while (i<8 && s[i] && s[i]!='.')
    dputc (s[i++]);
  j=i;
  while (s[i] && s[i]!='.')
    i++;
  if (s[i]=='.') {
    if (flags & DIR_show_dot) {
      dputc ('.');
    } else {
      while (j<9) {
	dputc (' ');
	j++;
      }
    }
    i++;
  }
  while (j<12 && s[i]) {
    dputc (s[i++]);
    j++;
  }
  while (j<12 && (flags & DIR_column)) {
    dputc (' ');
    j++;
  }

  return j;
}
// ====================================
// print file info
// ------------------------------------
static void ShowFileInfo (WIN32_FIND_DATA *dir, unsigned int flags) {
//int i, j;
WORD d, t;
  // validity check - not necessary
  //if (*dir->dir.name == '\x00' || *dir->dir.name == '\xE5') return;
  //if (!*dir->cAlternateFileName) return;
  if (!(flags & 0xFF)) flags |= /*DIR_show_short |*/ DIR_show_time | DIR_show_attr | DIR_show_size | DIR_show_LFN;
  // short name
  if (flags & DIR_show_short) {
    PrintFileName (dir, flags | DIR_force_short | DIR_column);
    dputc (' ');
  }
  // date & time
  if (flags & DIR_show_time) {
    FileTimeToDosDateTime (&dir->ftLastWriteTime , &d, &t);
    dprintf ("%2d.%2d.%d ", (d&0x1F), (d>>5)&0x0F, ((d>>9)&0x7F) + 1980);
    dprintf ("%02d:%02d:%02d ", t>>11, (t>>5)&0x3F, (t&0x1F)<<1);
  }
  // attributes
  if (flags & DIR_show_attr) {
    dputc ((dir->dwFileAttributes & FILE_ATTRIBUTE_ARCHIVE)?'A':'-');
    dputc ((dir->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)?'D':'-');
    //dputc ((dir->dwFileAttributes & FILE_ATTRIBUTE_VOLUME_ID)?'I':'-');
    dputc ((dir->dwFileAttributes & FILE_ATTRIBUTE_SYSTEM)?'S':'-');
    dputc ((dir->dwFileAttributes & FILE_ATTRIBUTE_HIDDEN)?'H':'-');
    dputc ((dir->dwFileAttributes & FILE_ATTRIBUTE_READONLY)?'R':'-');
  }
  // size
  if (flags & DIR_show_size) dprintf ("%10ld ", dir->nFileSizeLow);
  // internals
//  if (flags & DIR_show_internals)
//    dprintf ("cluster %04X%04X ", dir->dir.clusterH, dir->dir.clusterL);
  // LFN
  if (flags & DIR_show_LFN) PrintFileName (dir, DIR_show_dot);
//  if (*dir->dir.name == '\xE5') dprintf (" deleted");
  dputc ('\n');
}
// ====================================
// File Control Block functions
// ------------------------------------
// print name from FCB
// ------------------------------------
#define	FCB_dot		0x01
#define	FCB_drive	0x02
static void printFCB (FCB_type *fptr, char format) {
unsigned char i;
  if ((format & FCB_drive) && fptr->drive) {
    BDOS (console_output, fptr->drive-1+'A');
    BDOS (console_output, ':');
    if (!(format & FCB_dot))
      BDOS (console_output, ' ');
  }
  for (i=0; i<11; i++) {
    if (fptr->name[i]==' ' && (format & FCB_dot)) {
      if (i>=8) break;
      i=7;
      continue;
    }
    if (i==8)
      BDOS (console_output, (format & FCB_dot)?'.':' ');
    BDOS (console_output, fptr->name[i] & 0x7F);
  }
}
// ------------------------------------
// fill in name or extension field
// ------------------------------------
static int fillNameExt (char *ne, char *buf, unsigned char len) {
char *s;
  s=buf;
  //while (*buf>0 && *buf<=' ') buf++;
  while (len) {
    if (!*buf || *buf==' ' || *buf=='=' || *buf=='.' || *buf==':' || *buf==';' || *buf=='<' || *buf=='>')
      break;
    if (*buf=='*') {
      *ne++='?';
    } else {
      *ne++=*buf;
      buf++;
    }
    len--;
  }
  if (*buf=='*') buf++;
  while (len) {
    *ne++=' ';
    len--;
  }
  while (*buf && *buf!=' ' && *buf!='=' && *buf!='.' && *buf!=':' && *buf!=';' && *buf!='<' && *buf!='>')
    buf++;
  //while (*buf>0 && *buf<=' ') buf++;
  return buf-s;
}
// ------------------------------------
// fill whole FCB
// ------------------------------------
int fillFCB (FCB_type *fptr, char *buf) {
char *s;
unsigned char c;
  memset (fptr, 0, 16);
  s=buf;
  while (*buf>0 && *buf<=' ') buf++;
  if (*buf>='A' && *buf<='P' && buf[1]==':') {
    fptr->drive = *buf-'A'+1;
    buf+=2;
  }
  buf += fillNameExt (fptr->name, buf, 8);
  c=0;
  if (*buf=='.')
    buf++;
  buf += fillNameExt (fptr->ext, buf, 3);
//  printFCB (fptr, FCB_dot | FCB_drive); BDOS (console_output, '\n');
  return buf-s;
}
// ------------------------------------
// check for wildcards
// ------------------------------------
static unsigned char ambiguousFCB (FCB_type *fptr) {
unsigned char i;
  if (*fptr->name==' ' && *fptr->ext==' ') return 1;
  for (i=0; i<11; i++)
    if (fptr->name[i]=='?') return 1;
  return 0;
}
// ====================================
// get number from string
// ------------------------------------
int getnum (char *buf, int *val) {
char *s;
char sign;
  s=buf;
  sign=0;
  while (*buf>0 && *buf<=' ') buf++;
  //*val=0x80000000;
  *val=0;
  if (*buf=='-') {sign=1; buf++;}
  while (*buf>='0' && *buf<='9') {
    *val *= 10;
    *val += *buf-'0';
    buf++;
  }
  if (sign) *val = -*val;
  return buf-s;
}
// ====================================
// Console command processor
// ------------------------------------
char BDOS_verbose = 0;
extern char *ae_name;
void CCP (void) {
// command line variables
char *cmd;
unsigned char cmdp;
int i, cmdi;
// file system variables
FILE *f;
WIN32_FIND_DATA dir;
HANDLE hndl;
char path[1024];
//char my_LFN[LFN_path_len];
static unsigned char BDOS_flags;
unsigned short sval;
FILE *ae_f;

static unsigned long _CYC, _TIM = 0;
unsigned long cc, tc;
long td;
long cyc1ms, tim1ms, tim10ms;
unsigned long dt;
LARGE_INTEGER tt;
char SYSCALL;

  if (BDOS_flags&0xFE) BDOS_flags=0;
  // otevreni autoexec/$$$
  ae_f=NULL;
  if (ae_name)
    if (*ae_name)
      ae_f=fopen (ae_name, "rt");
  if (ae_f!=NULL)
    dprintf ("Executing %c:%s\n", 'A'+(curDisk & 0x0F), ae_name);

  // Console Command Processor
  cmdi = -1;
  do {
    // tisk promptu
    if (curpos) StartNewLine (0);
    BDOS (console_output, 'A' + (curDisk & 0x0F));
    if (curDisk > 0x9F)
      BDOS (console_output, '0' + ((curDisk >> 4)/10));
    if (curDisk > 0x0F)
      BDOS (console_output, '0' + ((curDisk >> 4)%10));
    BDOS (console_output, '>');
    // cteni prikazove radky
    CCPbuf[0]=127;
    CCPbuf[1]=0;
    if (ae_f) {
      // cteni ze SUBMIT file
      CCPbuf[2]=0;
      fgets ((char*)&CCPbuf[2], 126, ae_f);
      // smazani ridicich znaku na konci
      for (i=2; CCPbuf[i] && CCPbuf[i]!='\r' && CCPbuf[i]!='\n'; i++);
      CCPbuf[1]=i-2;
      CCPbuf[i]=0;
      // tisk radky
      dprintf ("%s", &CCPbuf[2]);
      if (feof (ae_f)) {
        fclose (ae_f);
	ae_f=NULL;
	ae_name=NULL;
/**
	if (*ae_name=='$') {
	  // po pouziti se musi soubor smazat
	  //FileDelete (ae_name);
	} else {
	  strcpy (ae_name, "$$$.SUB");
	}
**/
      }
    } else {
      // cteni z terminalu
      BDOS (read_console_buffer, CCPbuf_offs);
    }
    StartNewLine (0);
    // v originale se jeste kontroluje, zda se prikazy nectou z SUBMIT file...

    // prevod bufferu na ASCIIZ
    cmd = (char*)&CCPbuf[2];
    cmd[CCPbuf[1]]=0;
    while (*cmd>0 && *cmd<=' ') cmd++;
    if (!*cmd) continue;
    // prevod na velka pismena
    str_up (cmd);

    // nastaveni bufferu pro nektere operace
    BDOS (setDMA, defaultBUF_offs);

    // zpracovani prikazove radky
    cmdi = FindCommand (cmd);
    //dprintf ("\nCommand (%d): \"%s\"\n", cmdi, cmd);
    if (cmdi>=0) {
      cmd += strlen (commands[cmdi].cmd);
      while (*cmd>0 && *cmd<=' ') cmd++;
      cmdi=commands[cmdi].code;
    }
    switch (cmdi) {
      // ==============================
      // CP/M standard commands
      // ------------------------------
      // directory listing: DIR [<disk>][<name>.<ext>]
      // ------------------------------
      case cpm_dir:
        fillFCB (&defaultFCB, cmd);
	if (*defaultFCB.name==' ' && *defaultFCB.ext==' ') for (cmdp=0; cmdp<11; cmdp++) defaultFCB.name[cmdp]='?';
        if (BDOS (search_first, defaultFCB_offs) == 0xFF) break;
        do {
	  if (!curpos)
	    BDOS (console_output, 'A'+(defaultFCB.drive?(defaultFCB.drive-1):(curDisk&0x0F)));
	  BDOS (console_output, ':');
	  BDOS (console_output, ' ');
	  printFCB ((FCB_type*)defaultBUF, 0);
	  BDOS (console_output, ' ');
	  if (curpos >= 60)
	    BDOS (console_output, '\n');
	} while (BDOS (search_next, 0) != 0xFF);
	break;
      // ------------------------------
      // user selection: USER <user_num>
      // ------------------------------
      case cpm_user:
        getnum (cmd, &i);
	dprintf ("User: %d\n", i);
	if (i<0 || i>15) {
	  // error...
	  break;
	}
	BDOS (get_set_user, i);
        break;
      // ------------------------------
      // file typing on console: TYPE [<disk>]<name>.<ext>
      // ------------------------------
      case cpm_type:
        fillFCB (&defaultFCB, cmd);
	if (ambiguousFCB (&defaultFCB)) {
	  //unknownFile (&defaultFCB);
	  break;
	}
	if (3<(unsigned char)BDOS (open_file, defaultFCB_offs)) {
	  //error opening file
	  break;
	}
	BDOS (setDMA, defaultBUF_offs);
	while (!BDOS (read_file, defaultFCB_offs)) {
	  for (i=0; i<128; i++) {
	    if (defaultBUF[i]=='\x1A')
	      break;
	    BDOS (console_output, defaultBUF[i]);
	  }
//	  if (i<128 && defaultBUF[i]=='\x1A')
//	    break;
	}
	BDOS (close_file, defaultFCB_offs);
        break;
      // ------------------------------
      // run external command or disk select
      // ------------------------------
      default:
        cmd += fillFCB (&defaultFCB, cmd);
        if (*defaultFCB.name==' ' && *defaultFCB.ext==' ') {
	  BDOS (select_disk, defaultFCB.drive-1);
	  break;
	}
	if (ambiguousFCB (&defaultFCB) || *defaultFCB.ext!=' ') {
	  printFCB (&defaultFCB, FCB_drive | FCB_dot);
	  BDOS (console_output, '?');
	  BDOS (console_output, '\n');
	  break;
	}
	defaultFCB.ext[0]='C'; defaultFCB.ext[1]='O'; defaultFCB.ext[2]='M';
	// load TPA program
	if (3<(unsigned char)BDOS (open_file, defaultFCB_offs)) {
	  dprintf ("Error opening file!\n");
	  break;
	}
	i=0;
	do {
	  if (&TPA_start[i] >= TPA_top) break;
	  BDOS (setDMA, TPA_offs+i);
	  i+=128;
	} while (!BDOS (read_file, defaultFCB_offs));
	BDOS (close_file, defaultFCB_offs);
	if (&TPA_start[i] >= TPA_top) {
	  dprintf ("Program too large!\n");
	  break;
	}
	// run TPA program
      case cmd_tpa:
	// copy rest of command line to defaultBUF
	while (*cmd>0 && *cmd<=' ') cmd++;
	strcpy ((char*)&defaultBUF[1], cmd);
	*defaultBUF = strlen (cmd);
	// fill all two default FCBs
	cmd += fillFCB (&defaultFCB, cmd);
        while ((*cmd>0 && *cmd<=' ') || *cmd=='=' || *cmd=='.' || *cmd==':' || *cmd==';' || *cmd=='<' || *cmd=='>')
	  cmd++;
	cmd += fillFCB (&defaultFCB2, cmd);
        // eat LF from console
	for (i=50; i; i--) {
	  if (BDOS (console_direct, 0xFE)) {
	    BDOS (console_direct, 0xFF);
	    break;
	  }
	  delay (1);
	}
	// run TPA program
	BDOS_verbose = BDOS_flags;
	// inicializace prostredi simulatoru
	// ulozeni navratove adresy (WBOOT)
	CPM_RAM[SP_offs-1]=wboot_vec>>8;
	CPM_RAM[SP_offs-2]=wboot_vec&0xFF;
	Z80state.sp = SP_offs-2;
	Z80state.pc = TPA_offs;
	_CYC = _TIM = 0;
	cc = tc = td = 0;
	Z80init ();
	QueryPerformanceFrequency (&tt);
	tim10ms = (tt.u.LowPart+50)/100;
	tim1ms = (tt.u.LowPart+500)/1000;
	cyc1ms = freq;
	BREAK_request=0;
	do {
	  do {
	    // priprava prostredi simulatoru pro dalsi spusteni
	    Z80state.tstates = 0;
	    Z80state.event_next_event = cyc1ms; // pocet taktu do preruseni programu
	    // mereni skutecne doby simulace
	    // GetTickCount i GetSystemTime maji rozliseni (10..11)ms
	    QueryPerformanceCounter (&tt);
	    dt = tt.u.LowPart;
	    // spusteni simulace
	    Z80sim (NULL, (unsigned char*)&Z80state, NULL);
	    // kontrola rychlosti behu
	    QueryPerformanceCounter (&tt);
	    SYSCALL = (Z80state.tstates>100000000);
	    if (Z80state.tstates>100000000) Z80state.tstates-=100000000;
	    _CYC += Z80state.tstates;
	    cc += Z80state.tstates;
	    tc += tt.u.LowPart - dt;
	    dt = tt.u.LowPart;
	    // sledovani vykonnostnich ukazatelu a zpomalovani behu
	    if (cc >= cyc1ms) {
	      cc -= cyc1ms;
	      if (tc < tim1ms) {
		// kontrola preruseni programu
		dkbhit ();
		if (BREAK_request) break;
		// podle rezimu realna/maximalni rychlost musim pocitat dobu behu v [ms]!!!
		if (!(GetKeyState (VK_SCROLL) & 0x01)) {
		  td += tim1ms - tc;
		  tc = 0;
		  _TIM ++;
		  // zpomalovani behu
		  while (td > tim10ms) {
		    Sleep (10);
		    QueryPerformanceCounter (&tt);
		    td -= tt.u.LowPart - dt;
		    //if (td<0) dprintf ("!<0!");
		  }
		}
	      } else {
		while (tc > tim1ms) {
		  tc -= tim1ms;
		  _TIM ++;
		}
	      }
	    }
	  } while (!SYSCALL && !BREAK_request);
	  if (BREAK_request)
	    longjmp (BOOT_jmp, 0);
	  // volani systemovych funkci
	  if (Z80state.pc >= bdos_vec) {
	    switch (Z80state.pc-1) {
	    case bdos_trap: // BDOS
	      sval = BDOS (Z80state.c, (Z80state.d<<8) | Z80state.e);
	      if (BDOS_verbose) {
		dprintf ("BDOS(%02X, %02X%02X)=%04X from %02X%02X after %ld states\n", Z80state.c, Z80state.d, Z80state.e, sval, CPM_RAM[Z80state.sp+1], CPM_RAM[Z80state.sp], Z80state.tstates);
		//Sleep (100);
	      }
	      Z80state.l = Z80state.a = sval&0xFF;
	      Z80state.h = Z80state.b = sval>>8;
	      break;
	    case boot_trap: // BOOT
	      dprintf ("\nBIOS BOOT\n");
	      longjmp (BOOT_jmp, 0);
	    case boot_trap-2: // WBOOT
	      dprintf ("\nBIOS WBOOT\n");
	      longjmp (WBOOT_jmp, 0);
	    case boot_trap-4: // CONST
	      Z80state.a = CONSTAT ();
	      break;
	    case boot_trap-6: // CONIN
	      Z80state.a = CONIN ();
	      break;
	    case boot_trap-8: // CONOUT
	      CONOUT (Z80state.c);
	      break;
	    case boot_trap-10: // LIST
	      LIST (Z80state.c);
	      break;
	    case boot_trap-12: // PUNCH
	      PUNCH (Z80state.c);
	      break;
	    case boot_trap-14: // READER
	      Z80state.a = READER ();
	      break;
	    case boot_trap-16: // HOME
	      HOME ();
	      break;
	    case boot_trap-18: // SELDSK
	      sval = SELDSK (Z80state.c);
	      Z80state.l = sval&0xFF;
	      Z80state.h = sval>>8;
	      break;
	    case boot_trap-20: // SETTRK
	      SETTRK ((Z80state.b<<8) | Z80state.c);
	      break;
	    case boot_trap-22: // SETSEC
	      SETSEC ((Z80state.b<<8) | Z80state.c);
	      break;
	    case boot_trap-24: // SETDMA
	      SETDMA ((Z80state.b<<8) | Z80state.c);
	      break;
	    case boot_trap-26: // READ
	      Z80state.a = READ ();
	      break;
	    case boot_trap-28: // WRITE
	      Z80state.a = WRITE ();
	      break;
	    case boot_trap-30: // LISTST
	      Z80state.a = LISTST ();
	      break;
	    case boot_trap-32: // SECTRAN
	      sval = SECTRN ((Z80state.b<<8) | Z80state.c, (Z80state.d<<8) | Z80state.e);
	      Z80state.l = sval&0xFF;
	      Z80state.h = sval>>8;
	      break;
	    default:
	      // check if stop was caused by HALT or by reaching max Tstates
	      //if (Z80state.tstates>=Z80state.event_next_event) break; // happens sometimes... just one instruction before HALT...
	      dprintf ("Illegal BIOS/BDOS call! (%04X)\n", Z80state.pc);
	      if (Z80state.event_next_event!=1) // single-step???
		longjmp (WBOOT_jmp, 0);
	      break;
	    }
	  }
	} while (Z80state.pc-1 != boot_trap-2 && (Z80state.pc-1 != bdos_trap || Z80state.c != 0));
	BDOS_verbose = 0;
	// protoze by se ted mel znovu-spustit CCP, tak jde pres WBOOT, coz je jina moznost navratu do CCP
	longjmp (WBOOT_jmp, 0);
      // ------------------------------
      // real frequency simulation
      // ------------------------------
      case cmd_freq:
	if (*cmd) {
	  getnum (cmd, &freq);
	}
	dprintf ("Requested frequency: %dkHz\n", freq);
	if (_TIM)
	  dprintf ("Last simulated frequency: %ldkHz (%ld cycles in %ld ms)\n", _CYC/_TIM, _CYC, _TIM);
	break;
      // ------------------------------
      // BDOS flags
      // ------------------------------
      case cmd_bflags:
        getnum (cmd, &i);
        BDOS_flags = i;
        break;
      // ------------------------------
      // file erase: ERA [<disk>]<name>.<ext>
      // ------------------------------
      case cpm_era:
      // ------------------------------
      // file rename: REN [<disk>]<name>.<ext>=[<disk>]<name>.<ext>
      // ------------------------------
      case cpm_ren:
      // ------------------------------
      // TPA block save: SAVE <block_cnt> [<disk>]<name>.<ext>
      // ------------------------------
      case cpm_save:
	break;
      // ==============================
      // enhanced commands
      // ------------------------------
      // list all internal command
      // ------------------------------
      case cmd_help:
        dputs ("Available commands:\n");
        for (cmdp=0; cmdp<cmd_num; cmdp++)
          dprintf ("%s\n", commands[cmdp].cmd);
        break;
      // ------------------------------
      // reset CP/M simulator
      // ------------------------------
      case cmd_reset:
        // wait for LF character
        delay (100);
        // reset
	longjmp (BOOT_jmp, 0);
      // ------------------------------
      // change directory (for selected disk only)
      // ------------------------------
      case cmd_cd:
        strcpy (path, cmd);
        // set or get current working directory
        if (!SetCurrentDirectory (path)) {
          dputs ("Path not found!\n");
          break;
        }
        if (!*cmd) {
	  GetCurrentDirectory (1024, path);
          // print current working directory
          //dprintf ("%s", my_LFN);
          dputs (path);
	  dputc ('\n');
        }
        break;
      // ------------------------------
      // list all FAT16/32 directory entries
      // ------------------------------
      case cmd_dir:
        // printing directory list
        strcpy (path, cmd);
        if (!*cmd)
	  GetCurrentDirectory (1024, path);
        dprintf ("List of directory %s\n", path);
	strcat (path, "\\*.*");
	hndl = FindFirstFile (path, &dir);
        if (hndl==INVALID_HANDLE_VALUE) {
          dputs ("Error opening selected directory!\n");
          break;
        }
	do {
          //ShowFileInfo (&dir, DIR_show_short | DIR_show_time | DIR_show_size | DIR_show_attr | DIR_show_LFN | DIR_show_internals);
          ShowFileInfo (&dir, 0);
        } while (FindNextFile (hndl, &dir));
        FindClose (hndl);
        break;
      // ------------------------------
      // print selected file on terminal
      // ------------------------------
      case cmd_type:
        if (!*cmd) {
          dputs ("Wrong command syntax!\n");
          break;
        }
        f = fopen (cmd, "rt");
        if (!f) {
          dputs ("File not found!\n");
          break;
        }
        while (!feof (f)) {
          dputc (fgetc (f));
        }
        fclose (f);
        break;
      case cmd_exit:
	break;
    }
  } while (cmdi != cmd_exit);
  // pokud zadam prikaz quit nebo exit, tak to musi tady skoncit
}
