/* Emulation of the Z80 CPU with hooks into the other parts of z81.
 * Copyright (C) 1994 Ian Collier.
 * z81 changes (C) 1995-2001 Russell Marks.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _WIN32
// PalmOS headers
#include <PceNativeCall.h>
#include <Standalone.h>

// application specific or standard headers
#include "string.h"
#include "endian.h"
#include "armlet-ids.h"
#else

// replacement headers for x86 compiling
#include "PalmOS\PceNativeCall.h"
#include "PalmOS\endian_x86.h"
#include <string.h>
#endif

#define Z80_INTERNALS
#include "z80.h"

#define parity(a) (partable[a])

#ifndef _WIN32
STANDALONE_CODE_RESOURCE_ID(armz80_id);
#endif

unsigned char in (int h, int l);
unsigned char out (int h, int l, int a);

void Z80sim (const void *emulStateP, unsigned char *userData68KP, Call68KFuncType *call68KFuncP) {
// pamet procesoru
unsigned char *memram;
// registry procesoru
unsigned char a, f, b, c, d, e, h, l;
unsigned char r, a1, f1, b1, c1, d1, e1, h1, l1, i, iff1, iff2, im;
unsigned short pc;
unsigned short ix, iy, sp;
// vnitrni stavy simulatoru
unsigned int radjust;
unsigned char ixoriy, new_ixoriy;
unsigned char op;
unsigned char intsample;
int cc;
int intrq;
// pomocne promenne
unsigned int tstates, tsmax;
unsigned char partable[256];
int n;
/*
unsigned char *sc;
unsigned int counter;
unsigned char out254;
unsigned char keyb[9];
*/
  // vypocet pomocne tabulky parity
  for (n = 0; n < 256; n += 16) {
    int j, l;
    l = 0x9669 & (1 << (n >> 4)) ? 0x9669 : 0x6996;
    for (j = 0; j < 16; j++) {
      partable[n + j] = 4 * (l & 1);
      l >>= 1;
    }
  }

  // nacteni dat z globalni promenne
  // adresa RAM
  memram = (unsigned char *)ReadUnaligned32(&(((Z80state_type*)userData68KP)->ram));
  // stav citace instrukcnich cyklu
  tstates = ReadUnaligned32(&(((Z80state_type*)userData68KP)->tstates));
  // maximalni cas citace instrukcnich cyklu (pak se vrati do hlavniho programu)
  tsmax = ReadUnaligned32(&(((Z80state_type*)userData68KP)->event_next_event));
  // registry procesoru
  a = ((Z80state_type*)userData68KP)->a;
  f = ((Z80state_type*)userData68KP)->f;
  b = ((Z80state_type*)userData68KP)->b;
  c = ((Z80state_type*)userData68KP)->c;
  d = ((Z80state_type*)userData68KP)->d;
  e = ((Z80state_type*)userData68KP)->e;
  h = ((Z80state_type*)userData68KP)->h;
  l = ((Z80state_type*)userData68KP)->l;
  r = ((Z80state_type*)userData68KP)->r | 0x80;
  a1 = ((Z80state_type*)userData68KP)->a1;
  f1 = ((Z80state_type*)userData68KP)->f1;
  b1 = ((Z80state_type*)userData68KP)->b1;
  c1 = ((Z80state_type*)userData68KP)->c1;
  d1 = ((Z80state_type*)userData68KP)->d1;
  e1 = ((Z80state_type*)userData68KP)->e1;
  h1 = ((Z80state_type*)userData68KP)->h1;
  l1 = ((Z80state_type*)userData68KP)->l1;
  i = ((Z80state_type*)userData68KP)->i;
  iff1 = ((Z80state_type*)userData68KP)->iff1;
  iff2 = ((Z80state_type*)userData68KP)->iff2;
  im = ((Z80state_type*)userData68KP)->im;
  pc = ReadUnaligned16(&(((Z80state_type*)userData68KP)->pc));
  ix = ReadUnaligned16(&(((Z80state_type*)userData68KP)->ix));
  iy = ReadUnaligned16(&(((Z80state_type*)userData68KP)->iy));
  sp = ReadUnaligned16(&(((Z80state_type*)userData68KP)->sp));

  ixoriy = new_ixoriy = 0;
  tstates = radjust = 0;
  cc = 1;
  intrq = 0;
  while(1) {
    ixoriy = new_ixoriy;
    new_ixoriy = 0;
    intsample = 1;
    if(!intrq || !cc || !iff1) {
      op = fetch(pc);
      pc++;
    } else {
      // preruseni
#if 0
      op = 0xff;
      cc = 0;
#else
      if(im!=2) {
	// IM0 nebo IM1
	op = 0xff;			/* rst 38 */
	cc = 0;
      } else  {
	// IM2
	int a;
	sp-=2;
	iff1=0;
	store2(sp, pc);
	a=((unsigned int)i<<8)+255;
	pc=fetch2(a);
	tstates+=19;
	op=fetch(pc);
	pc++;
	cc=0;
      }
#endif
    }
    radjust++;

    switch (op) {
#include "z80ops.c"
    }
    // test na konec casoveho useku simulace
    if((tstates > tsmax) && !new_ixoriy) {
      // ulozeni registru
      ((Z80state_type*)userData68KP)->a = a;
      ((Z80state_type*)userData68KP)->f = f;
      ((Z80state_type*)userData68KP)->b = b;
      ((Z80state_type*)userData68KP)->c = c;
      ((Z80state_type*)userData68KP)->d = d;
      ((Z80state_type*)userData68KP)->e = e;
      ((Z80state_type*)userData68KP)->h = h;
      ((Z80state_type*)userData68KP)->l = l;
      ((Z80state_type*)userData68KP)->r = r;
      ((Z80state_type*)userData68KP)->a1 = a1;
      ((Z80state_type*)userData68KP)->f1 = f1;
      ((Z80state_type*)userData68KP)->b1 = b1;
      ((Z80state_type*)userData68KP)->c1 = c1;
      ((Z80state_type*)userData68KP)->d1 = d1;
      ((Z80state_type*)userData68KP)->e1 = e1;
      ((Z80state_type*)userData68KP)->h1 = h1;
      ((Z80state_type*)userData68KP)->l1 = l1;
      ((Z80state_type*)userData68KP)->i = i;
      ((Z80state_type*)userData68KP)->iff1 = iff1;
      ((Z80state_type*)userData68KP)->iff2 = iff2;
      ((Z80state_type*)userData68KP)->im = im;
      WriteUnaligned16 (&(((Z80state_type*)userData68KP)->pc), pc);
      WriteUnaligned16 (&(((Z80state_type*)userData68KP)->ix), ix);
      WriteUnaligned16 (&(((Z80state_type*)userData68KP)->iy), iy);
      WriteUnaligned16 (&(((Z80state_type*)userData68KP)->sp), sp);
      // ulozeni stavu citace instrukcnich cyklu
      WriteUnaligned32 (&(((Z80state_type*)userData68KP)->tstates), tstates);

      return;
    }
  }
}
// vstup z portu
unsigned char in (int h, int l) {
  return 255;
}
// vystup na port
unsigned char out (int h, int l, int a) {
  return 0;
}
